import os
from pathlib import Path

from loguru import logger

STUB_TEMPLATE = """from typing import Any
from .config import StelaOptions
from .main import StelaMain

class Stela:
    _stela_options: StelaOptions
    _stela_data: StelaMain

    current_environment: str
    default_environment: str
    def get(self, var_name: str, raise_on_missing: bool = True) -> Any: ...
    def get_or_default(self, var_name: str, default: Any) -> Any: ...
    def list(self) -> list[str]: ...

    # Dynamic attributes
    {all_envs}


env: Stela

"""

STUB_MODULE = "__init__.pyi"


def create_stela_stub(settings) -> tuple[bool, str]:
    try:
        import stela

        # Get os.environ only for the keys that are in fact environment keys
        filtered_environ = {
            k: v for k, v in os.environ.items() if not k.startswith("_") and k.isupper()
        }

        all_env_info = {**filtered_environ, **settings}

        content = STUB_TEMPLATE.format(
            all_envs="\n    ".join(
                [f"{k}: {type(v).__name__}" for k, v in all_env_info.items()]
            )
        )
        stela_dir = os.path.dirname(os.path.abspath(stela.__file__))
        file_path = Path.cwd().joinpath(stela_dir, STUB_MODULE)
        if not file_path.exists():
            old_content = ""
        else:
            with open(file_path, "r") as f:
                old_content = f.read()
        if old_content != content:
            with open(file_path, "w") as f:
                f.write(content)
                message = f"Updated Stela Stub at: {file_path}"
            logger.success(message)
        else:
            message = f"Stela Stub is up to date at: {file_path}"
            logger.debug(message)
        return True, message
    except Exception as exc:
        message = f"Error while attempt to upsert Stela Stub: {exc}"
        logger.warning(message)
        return False, message



import importlib
import os
from dataclasses import dataclass, field

from loguru import logger

from stela.config import StelaOptions
from stela.exceptions import StelaEnvironmentNotFoundError
from stela.parsers.dotenv import read_dotenv
from stela.utils import show_value


@dataclass
class StelaMain:
    options: StelaOptions
    settings: dict[str, any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Save original environment keys.

        This is needed to avoid overwriting true system environment variables
        but to override values that were previously created by dotenv files (like .env)
        when reading the same variable again, in new dotenv files
        (like .env.local or .env.production).
        """
        original_keys = os.environ.get("_ORIGINAL_ENVIRON_KEYS", "")
        if not original_keys:
            os.environ["_ORIGINAL_ENVIRON_KEYS"] = ",".join(sorted(os.environ.keys()))

    def log_current_data(self, origin: str) -> None:
        def log_dict(dict_key: str, dict_value: any, parents: list):
            if isinstance(dict_value, dict):
                for k, v in dict_value.items():
                    if dict_key not in parents:
                        parents.append(dict_key)
                    log_dict(k, v, parents)
            else:
                parents_key = (
                    parents + [dict_key] if dict_key not in parents else parents
                )
                logger.debug(
                    f"[{origin}] {'.'.join(parents_key)} = "
                    f"{show_value(dict_value, self.options.log_filtered_value)}"
                )

        for key, value in self.settings.items():
            log_dict(key, value, [])

    @property
    def environment(self) -> str:
        """Return Current Environment."""
        if not self.options.current_environment:
            raise StelaEnvironmentNotFoundError("Environment not found.")
        return self.options.current_environment

    def get_project_settings(self) -> None:
        env_data = self.read_env_files()

        module_path = ".".join(self.options.final_loader.split(".")[:-1])
        loader_fn = self.options.final_loader.split(".")[-1]
        module = importlib.import_module(module_path)
        self.settings = getattr(module, loader_fn)(
            options=self.options, env_data=env_data
        )
        # Preserve existing system env values; allow loader to override only dotenv-injected keys
        forbidden_keys = os.getenv("_ORIGINAL_ENVIRON_KEYS", "").split(",")
        for k, v in self.settings.items():
            if v is None:
                continue
            if k not in forbidden_keys:
                os.environ[k] = v
        if self.options.show_logs:
            logger.debug(
                f"Settings after running loader: "
                f"{[f'{k}={show_value(v, self.options.log_filtered_value)}' for k, v in self.settings.items()]}"
            )

    def read_env_files(self) -> dict[str, any]:
        settings = {}
        current_environment = self.options.current_environment
        dotenv_files = [
            self.options.env_file,
            f"{self.options.env_file}.local",
        ]
        if current_environment:
            dotenv_files += [
                f"{self.options.env_file}.{current_environment.lower()}",
                f"{self.options.env_file}.{current_environment.lower()}.local",
            ]
        if self.options.show_logs:
            logger.debug(f"Looking for dotenv files: {dotenv_files}")
        for file in dotenv_files:
            env_settings = read_dotenv(
                config_file_path=self.options.config_file_path,
                env_file=file,
                encoding=self.options.dotenv_encoding,
                verbose=self.options.warn_if_env_is_missing,
                show_logs=self.options.show_logs,
                filter_logs=self.options.log_filtered_value,
            )
            settings |= env_settings
        # Preserve existing system env values; allow loader to override only dotenv-injected keys
        forbidden_keys = os.getenv("_ORIGINAL_ENVIRON_KEYS", "").split(",")
        for k, v in settings.items():
            if v is None:
                continue
            if k not in forbidden_keys:
                os.environ[k] = v
        if self.options.show_logs:
            logger.debug(
                f"Settings after load all dotenv files: "
                f"{[f'{k}={show_value(v, self.options.log_filtered_value)}' for k, v in settings.items()]}"
            )
        return settings


def default_loader(options: StelaOptions, env_data: dict[str, any]) -> dict[str, any]:
    """Stela Default Loader.

    :param options: Stela Options
    :param env_data: Dict with environment data
    :return: Dict
    """
    from loguru import logger

    logger.info(
        f"Using Stela Default Loader. Current environment is: {options.current_environment}"
    )
    return env_data



import configparser
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

import toml


@dataclass
class StelaFileReader:
    """Read other files to load settings.

    Will be removed on 6.0
    """

    def load_ini(self, filepath: Path) -> Dict[Any, Any]:
        """Load INI files.

        :param filepath: Path instance
        :return: Dict
        """
        config = configparser.ConfigParser()
        config.read(filepath)
        ini_settings: Dict[Any, Any] = {}
        for main_key in config.keys():
            ini_settings[main_key] = {}
            for key in config[main_key].keys():
                ini_settings[main_key][key] = config[main_key][key]

        return ini_settings

    def load_toml(self, filepath: Path) -> Dict[Any, Any]:
        """Load TOML files.

        :param filepath: Path instance
        :return: Dict
        """
        config = toml.load(filepath)
        return config  # type: ignore



"""Import Stela Data.

If you need to reload data inside you application,
use the stela_reload function.

"""

import os
from ast import literal_eval
from typing import Any, List

from loguru import logger

from stela.config import StelaOptions
from stela.exceptions import StelaValueError
from stela.helpers.stub import create_stela_stub
from stela.main import StelaMain
from stela.utils import show_value

__version__ = "8.0.10"


def _get_stela() -> "Stela":
    stela_config = StelaOptions.get_config()

    stela_data = StelaMain(options=stela_config)
    stela_data.get_project_settings()

    # Enable or disable logs based on configuration
    # after the data is loaded
    if stela_config.show_logs:
        logger.enable("stela")
    else:
        logger.info("Disabling Stela logs.")
        logger.disable("stela")

    class Stela:
        __slots__ = (
            ["_locked"] + [str(k) for k in stela_data.settings.keys()]
            if stela_data.settings.keys()
            else ["_locked"]
        )
        _stela_options: StelaOptions = stela_config
        _stela_data: StelaMain = stela_data

        def _get_attributes(self, current_obj: object, data_dict: dict):
            for attr, value in data_dict.items():
                if isinstance(value, dict):

                    class StelaNestedObj:
                        __slots__ = [str(k) for k in value.keys()]

                    nested_obj = StelaNestedObj()
                    self._get_attributes(current_obj=nested_obj, data_dict=value)
                    setattr(current_obj, attr, nested_obj)
                else:
                    current_value = self._evaluate_value(attr, value)
                    setattr(current_obj, attr, current_value)

        @property
        def current_environment(self):
            return (
                self._stela_options.current_environment
                or self._stela_options.no_env_name
            )

        @property
        def default_environment(self):
            return (
                self._stela_options.default_environment
                or self._stela_options.no_env_name
            )

        def __init__(self, *args, **kwargs):
            self._locked = False
            super().__init__(*args, **kwargs)

            self._get_attributes(current_obj=self, data_dict=stela_data.settings)

        def lock(self):
            self._locked = True

        @staticmethod
        def _evaluate_value(item, value: Any) -> Any:
            if stela_config.evaluate_data and isinstance(value, str):
                try:
                    current_value = literal_eval(value)
                except Exception:
                    current_value = value
                else:
                    logger.debug(
                        f"Using evaluated value for: {item}. Type is: {type(current_value).__name__}"
                    )
                return current_value
            return value

        def __getattribute__(self, item):
            # return from os.environ if exists
            if item in os.environ:
                value = os.environ[item]
                logger.debug(
                    f"Using environment value: {item}={show_value(value, stela_config.log_filtered_value)}"
                )
                return self._evaluate_value(item, value)
            try:
                value = super().__getattribute__(item)
                if not item.startswith("_"):
                    logger.debug(
                        f"Using stela value: {item}={show_value(value, stela_config.log_filtered_value)}"
                    )
                return value
            except AttributeError as exc:
                if stela_config.raise_on_missing_variable:
                    raise StelaValueError(
                        f"Stela did not found value for {item}."
                    ) from exc
                logger.warning(f"Stela did not found value for {item}. Returning None.")
                return None

        def __setattr__(self, key, value):
            if key in self.__slots__ and hasattr(self, key):
                if key != "_locked" and self._locked:
                    raise StelaValueError(f"Attribute {key} is ready-only.")
            super().__setattr__(key, value)

        def get(self, var_name: str, raise_on_missing: bool = True):
            if raise_on_missing:
                try:
                    return getattr(self, var_name)
                except AttributeError as exc:
                    raise StelaValueError(
                        f"Stela did not found value for {var_name}."
                    ) from exc
            return getattr(self, var_name, None)

        def get_or_default(self, var_name: str, default: Any):
            return getattr(self, var_name, default)

        def list(self) -> List[str]:
            return [k for k in self._stela_data.settings.keys()]

    create_stela_stub(stela_data.settings)

    stela = Stela()
    stela.lock()
    return stela


env = _get_stela()



from asgiref.sync import sync_to_async


async def aread_env() -> "Stela":
    """Reload Stela configuration asynchronous.

    Not really needed for default configuration
    but in some cases, if you use a complex blocking custom loader,
    all operation will be wrapped in a coroutine here.

    """
    from stela.utils import read_env

    return await sync_to_async(read_env)()



from typing import Any, Dict, Tuple

from loguru import logger
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings, PydanticBaseSettingsSource

from stela.utils import read_env


def stela_env_settings(settings: BaseSettings) -> dict[str, Any]:
    """
    Read Stela Data from many dotenv files.
    """

    delimiter = settings.config.get("env_nested_delimiter") or ""
    log_settings = settings.config.get("env_nested_delimiter") or False

    # Ask Stela to read envs as per his configuration
    env = read_env()

    if delimiter:
        data = {
            ".".join(var.split(delimiter)).lower(): env.get(var) for var in env.list()
        }
    else:
        data = {var.lower(): env.get(var) for var in env.list()}

    if log_settings:
        logger.debug(f"Stela settings dict are: {data}")

    return data


class StelaConfigSettingsSource(PydanticBaseSettingsSource):
    """
    Stela Config Settings Source.
    """

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> Tuple[Any, str, bool]:
        stela_data = stela_env_settings(self)
        field_value = stela_data.get(field_name)
        return field_value, field_name, False

    def prepare_field_value(
        self, field_name: str, field: FieldInfo, value: Any, value_is_complex: bool
    ) -> Any:
        return value

    def __call__(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {}

        for field_name, field in self.settings_cls.model_fields.items():
            field_value, field_key, value_is_complex = self.get_field_value(
                field, field_name
            )
            field_value = self.prepare_field_value(
                field_name, field, field_value, value_is_complex
            )
            if field_value is not None:
                d[field_key] = field_value

        return d



"""Stela exceptions."""


class StelaEnvironmentNotFoundError(Exception):
    """Environment not found Error."""


class StelaFileTypeError(Exception):
    """File Type Error."""


class StelaValueError(AttributeError):
    """Stela Value Error."""



from pathlib import Path
from typing import Any, Dict, Optional

from dotenv import dotenv_values, find_dotenv

from stela.utils import show_value


def read_dotenv(
    config_file_path: str,
    env_file: str,
    verbose: bool,
    encoding: str,
    show_logs: bool = False,
    filter_logs: bool = True,
) -> Dict[Any, Any]:
    """Stela DotEnv Loader.

    :param config_file_path: relative path for config files
    :param env_file: dotenv file name
    :param verbose: warn if dotenv file is not found
    :param encoding: encoding for dotenv file
    :param show_logs: show logs for dotenv file reading
    :param filter_logs: Filter variable value in logs
    :return: Dict
    """
    from loguru import logger

    def look_for_file(current_path: Path, file_name: str) -> Optional[Path]:
        test_path = current_path.joinpath(file_name)
        if show_logs:
            logger.debug(f"Looking for file '{env_file}' in '{current_path}'")
        if test_path.exists():
            return test_path
        if str(current_path) in ["/", "\\"] or current_path.parent == current_path:
            return None
        return look_for_file(current_path.parent, file_name)

    path = Path.cwd().joinpath(config_file_path)
    env_path = look_for_file(path, env_file)
    if not env_path:
        if show_logs:
            logger.debug(f"File {env_file} not found starting at path {path}.")
        return {}
    logger.info(f"Reading file: {env_path}")  # always log the file being read
    dotenv_path = find_dotenv(str(env_path), usecwd=True)
    if not dotenv_path:
        return {}

    env_data = dotenv_values(
        dotenv_path=dotenv_path, verbose=verbose, encoding=encoding, interpolate=False
    )
    if env_data and show_logs:
        logger.debug(
            f"Filtered values from {dotenv_path}: {[f'{k}={show_value(v, filter_logs)}' for k, v in env_data.items()]}"
        )
    return env_data



"""Stela Initialization.

What we need here?

First of all, check if project already uses stela.

# Existing project (only)
1. Check if it needs to convert old keys.
2. If needed, convert old keys and create .env files.
3. Check if .gitignore exists. If not, create it. On this file, add the stela ignored files.

# New project
1. Check if .gitignore exists. If not, create it. On this file, add the stela ignored files.
2. Ask for stela env name, default is STELA_ENV.
3. Ask for default environment, default is development.

"""

import os
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import click
import tomlkit

from stela.config import StelaOptions

STELA_INI_FILE = ".stela"

PYPROJECT_TOML = "pyproject.toml"


@dataclass
class StelaInit:
    root_dir: str
    old_env_config_found: bool = False
    env_file: str = StelaOptions.env_file
    default_environment: Optional[str] = None
    dotenv_file_path: str = StelaOptions.config_file_path

    def upsert_gitignore(self) -> bool:
        file_exists = os.path.exists(".gitignore")
        if file_exists:
            old_content = open(".gitignore", "r").read()
        else:
            old_content = ""
            self.old_env_config_found = True
        new_content = deepcopy(old_content)
        if ".env\n" in new_content:
            new_content.replace(".env\n", "!.env  # Modified by Stela\n")
            self.old_env_config_found = True
        if "*.stela.back" not in new_content:
            new_content += "\n# Stela\n"
            new_content += "*.stela.back\n"
        if f"{self.env_file}.local" not in new_content:
            new_content += f"{self.env_file}.local\n"
        if f"{self.env_file}.*.local" not in new_content:
            new_content += f"{self.env_file}.*.local\n"
        if new_content != old_content:
            click.secho("Updating .gitignore file.", fg="green")
            with open(".gitignore", "w") as f:
                f.write(new_content)

    def add_stela_config(self, use_default):
        from stela.parsers.reader import StelaFileReader

        reader = StelaFileReader()
        stela_config = {}
        pyproject_doc = None
        pyproject_content = None
        if os.path.exists(STELA_INI_FILE):
            stela_config = reader.load_ini(STELA_INI_FILE)
        if not stela_config and os.path.exists(PYPROJECT_TOML):
            with open("pyproject.toml", "r") as f:
                pyproject_content = f.read()
            pyproject_doc = tomlkit.parse(pyproject_content)
            toml_settings = reader.load_toml(PYPROJECT_TOML)
            stela_config = toml_settings.get("tool", {}).get("stela", {})

        if use_default:
            new_stela_config = {
                "environment_variable_name": StelaOptions.environment_variable_name,
                "evaluate_data": StelaOptions.evaluate_data,
                "show_logs": StelaOptions.show_logs,
                "env_file": StelaOptions.env_file,
                "config_file_path": StelaOptions.config_file_path,
            }
        else:
            new_stela_config = {
                "environment_variable_name": click.prompt(
                    "Stela environment variable name",
                    default=stela_config.get(
                        "environment_variable_name",
                        StelaOptions.environment_variable_name,
                    ),
                ),
            }
            use_default_env = click.confirm("Use default environment?", default=False)
            if use_default_env:
                self.default_environment = click.prompt(
                    "Default environment",
                    default=stela_config.get("default_environment", "development"),
                )
                new_stela_config["default_environment"] = self.default_environment
            new_stela_config["evaluate_data"] = click.confirm(
                "Evaluate data from env files?",
                default=stela_config.get("evaluate_data", StelaOptions.evaluate_data),
            )
            new_stela_config["show_logs"] = click.confirm(
                "Show logs?",
                default=stela_config.get("show_logs", StelaOptions.show_logs),
            )
            if new_stela_config["show_logs"]:
                new_stela_config["log_filtered_value"] = click.confirm(
                    "Show filtered values in logs?",
                    default=stela_config.get(
                        "log_filtered_value", StelaOptions.log_filtered_value
                    ),
                )
            self.env_file = click.prompt(
                "Default env file",
                default=stela_config.get("env_file", StelaOptions.env_file),
            )
            new_stela_config["env_file"] = self.env_file
            self.dotenv_file_path = click.prompt(
                "Default relative path for dotenv files?",
                default=stela_config.get(
                    "config_file_path", StelaOptions.config_file_path
                ),
            )
            new_stela_config["config_file_path"] = self.dotenv_file_path

        if not use_default:
            use_toml = (
                click.confirm("Save stela config in pyproject.toml file?", default=True)
                if pyproject_content
                else False
            )
        else:
            use_toml = False

        if use_toml:
            pyproject_doc.setdefault("tool", {})["stela"] = new_stela_config
            with open(PYPROJECT_TOML, "w") as f:
                f.write(tomlkit.dumps(pyproject_doc))
                click.secho(
                    f"Updating {PYPROJECT_TOML} file with stela configuration", dim=True
                )
            if os.path.exists(STELA_INI_FILE):
                os.remove(STELA_INI_FILE)
                click.secho(f"Removing file {STELA_INI_FILE}", dim=True)
        else:
            with open(STELA_INI_FILE, "w") as f:
                f.write("[stela]\n")
                for k, v in new_stela_config.items():
                    f.write(f"{k} = {v}\n")
                click.secho(
                    f"Updating {STELA_INI_FILE} file with stela configuration", dim=True
                )
            if pyproject_content and "[tool.stela]" in pyproject_content:
                del pyproject_doc["tool"]["stela"]
                with open(PYPROJECT_TOML, "w") as f:
                    f.write(tomlkit.dumps(pyproject_doc))
                click.secho(
                    f"Removing stela configuration from  {PYPROJECT_TOML} file",
                    dim=True,
                )

    def run(self, use_default):
        if use_default:
            click.secho("Using default stela config options.", fg="green")
        self.add_stela_config(use_default)
        self.upsert_gitignore()
        self.create_env_files()
        click.secho("Stela successfully initialized.", fg="green", bold=True)

    def _create_env_file(self, filename, comment):
        path = os.path.join(Path.cwd(), self.dotenv_file_path)
        if not os.path.exists(path):
            os.makedirs(path)
        file_path = os.path.join(path, filename)
        file_exists = os.path.exists(file_path)
        if not file_exists:
            path_resolved = os.path.relpath(file_path, Path.cwd())
            click.secho(f"Creating {path_resolved}", dim=True)
            with open(file_path, "w") as f:
                f.write(f"# {comment}\n")

    def create_env_files(self):
        file_exists = os.path.exists(".env")
        if file_exists:
            if self.old_env_config_found:
                click.secho("Found old .env file. Renaming to .env.local", fg="yellow")
                os.rename(".env", ".env.local")
        else:
            self._create_env_file(
                self.env_file,
                "Add here your settings and fake secrets. "
                "You can commit this file.\n\n#MY_SECRET=fake_value",
            )
            self._create_env_file(
                f"{self.env_file}.local",
                "Add here your local settings and/or real secrets. "
                "DO NOT commit this file.\n\n#MY_SECRET=real_value",
            )
            if self.default_environment:
                self._create_env_file(
                    f"{self.env_file}.{self.default_environment}",
                    f"Add here your settings for the default environment. "
                    f"You can commit this file.\n\n#MY_{self.default_environment.upper()}_SECRET=fake_value",
                )
                self._create_env_file(
                    f"{self.env_file}.{self.default_environment}.local",
                    f"Add here your local settings for the default environment. "
                    f"DO NOT commit this file.\n\n#MY_{self.default_environment.upper()}_SECRET=real_value",
                )
